#include <glib.h>
#include <glib/gi18n.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "libbox.h"
#include "libbox-db.h"
#include "libbox-db-deps.h"
#include "libbox-db-utils.h"
#include "libbox-db-repos.h"
#include "libbox-parser.h"
#include "libinternal-utils.h"

static gchar* arg_install = NULL;
static gchar* arg_release = "1";
static gchar* arg_pkgname = NULL;
static gchar* arg_pkgsummary = NULL;
static gchar* source_dir = NULL;
static gchar* dest_dir = "/";

static gchar* arg_deps_r = NULL;
static gchar* arg_deps_br = NULL;
static gchar* arg_deps_p = NULL;
static gchar* arg_deps_bp = NULL;
static gchar* summary = "";
static gboolean arg_topology = FALSE;

static GOptionEntry entries[] = 
{
    { "install", 'i', 0, G_OPTION_ARG_STRING, &arg_install, "Install package using xml box file", NULL },
    { "topology", 't', 0, G_OPTION_ARG_NONE, &arg_topology, "Data for packages topology sort", NULL },
    { "source-dir", 's', 0, G_OPTION_ARG_STRING, &source_dir, "Specify directory for files", NULL },
    { "root", 0, 0, G_OPTION_ARG_STRING, &dest_dir, "Installation dir, defaults /", NULL },
    { "release", 'r', 0, G_OPTION_ARG_STRING, &arg_release, "Release number", NULL },
    { "pkg-name", 'p', 0, G_OPTION_ARG_STRING, &arg_pkgname, "Package name", NULL },
    { "summary", 'p', 0, G_OPTION_ARG_STRING, &arg_pkgsummary, "Package summary", NULL },
    { "deps-r", 0, 0, G_OPTION_ARG_STRING, &arg_deps_r, "Deps: require", NULL },
    { "deps-br", 0, 0, G_OPTION_ARG_STRING, &arg_deps_br, "Deps: build require", NULL },
    { "deps-p", 0, 0, G_OPTION_ARG_STRING, &arg_deps_p, "Deps: provide", NULL },
    { "deps-bp", 0, 0, G_OPTION_ARG_STRING, &arg_deps_bp, "Deps: build provide", NULL },
    { NULL }
};

int main (int argc, char *argv[])
{
    GError *error = NULL;
    GOptionContext *context;
    gint i;

    GList *files = NULL;
    GList *files0 = NULL;
    GList *li = NULL;
    GList *li0 = NULL;
    gchar *full_path = NULL;
    sqlite3 *db = NULL;
    gint package_id;
    BoxDocument *document = NULL;

    context = g_option_context_new ("box-db-manager");
    g_option_context_add_main_entries (context, entries, NULL);
    g_option_context_set_help_enabled(context, TRUE);
    g_option_group_set_error_hook(g_option_context_get_main_group(context), (GOptionErrorFunc)goption_error_func);
    g_option_context_parse (context, &argc, &argv, &error);

    g_option_context_free(context);

    db = box_db_open(dest_dir);
    box_db_create_empty(db);

    if (arg_install)
    {
        if (!source_dir)
            g_error("Please specify --source-dir parameter");

        document =  box_document_new(arg_install);

        box_db_transaction_begin(db);
        if (arg_pkgname)
            files0 = box_db_get_file_list(db, arg_pkgname);
        else
            files0 = box_db_get_file_list(db, document->header->name);

        error = NULL;
        if (arg_pkgsummary)
        {
            if (!g_file_get_contents(arg_pkgsummary, &summary, NULL, &error))
                summary = g_strdup("");
            else {
                for (i=0; i<strlen(summary); i++)
                {
                    if (summary[i] == 0x0a)
                        summary[i] = 0x20;
                    g_strstrip(summary);
                }
            }
        }

        if (arg_pkgname)
        {
            files = box_db_install(db, source_dir, "/", arg_pkgname, document->header->version, atoi(arg_release), summary);
            package_id = box_db_get_package_id(db, arg_pkgname);
        } else
        {
            files = box_db_install(db, source_dir, "/", document->header->name, document->header->version, atoi(arg_release), summary);
            package_id = box_db_get_package_id(db, document->header->name);
        }
        g_free(summary);

        for (li0 = files0; li0 != NULL; li0 = li0->next)
        {
            gboolean exist = FALSE;

            for (li = files; li != NULL; li = li->next)
            {
                if (strcmp((gchar *)li0->data, (gchar *)li->data) == 0)
                    exist = TRUE;
            }
            if (!exist)
            {
                full_path = g_build_filename(dest_dir, (gchar *)li0->data, NULL);
                printf("remove: %s\n", full_path);
                g_free(full_path);
            }
        }
        box_document_free(document);

        for (li = files; li != NULL; li = li->next)
        {
            full_path = g_build_filename(dest_dir, (gchar *)li->data, NULL);
            printf("install: %s\n", full_path);
            g_free(full_path);   
        }

        box_db_file_list_free(files0);

        if (arg_deps_r)
            box_db_install_requires(db, package_id, arg_deps_r);
        if (arg_deps_br)
            box_db_install_build_requires(db, package_id, arg_deps_br);
        if (arg_deps_p)
            box_db_install_provides(db, package_id, arg_deps_p);
        if (arg_deps_bp)
            box_db_install_build_provides(db, package_id, arg_deps_bp);

        box_db_transaction_end(db);
    }

    box_db_close(db);

    if (arg_topology)
    {
        GList *packages;
        gchar *package;
        GList *deps;
        GList *lidep;
        PackageSearch *ps;
       
        box_db_open(dest_dir);
        box_db_attach_repos(db, box_get_repos_dir(dest_dir));
        box_db_repos_init(db);

        packages = box_db_repos_get_packages(db);
        for (li = packages; li != NULL; li = li->next)
        {
            package = (gchar*) li->data;
            g_assert(package != NULL);

            deps = box_db_repos_get_depends(db, package);
            if (deps)
            {
                for (lidep = deps; lidep != NULL; lidep = lidep->next)
                {
                    ps = (PackageSearch*) lidep->data;
                    g_assert(ps != NULL);
                    g_assert(ps->package != NULL);
                    printf("%s\t%s\n", package, ps->package);
                }
                box_db_repos_package_list_free(deps);
            }
        }
        box_db_detach_repos(db);
        box_db_close(db);
    }

    return 0;
}
