#include <glib.h>
#include <glib/gi18n.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>

#include "libbox.h"
#include "libbox-db.h"
#include "libbox-db-deps.h"
#include "libbox-db-utils.h"
#include "libbox-db-repos.h"
#include "libbox-parser.h"
#include "libinternal-utils.h"

/*
TODO:
  - python imports
  - perl imports
  - ruby imports
*/

static gchar* arg_mode = "file";
static gchar* dest_dir = "/";

static GOptionEntry entries[] = 
{
    { "mode", 'm', 0, G_OPTION_ARG_STRING, &arg_mode, "Select not-found mode. One from: file, command, man, pkgconfig, header.", "MODE" },
    { "root", 0, 0, G_OPTION_ARG_STRING, &dest_dir, "Installation dir, defaults /", NULL },
    { NULL }
};


static void print_install_command(const gchar *path, const gchar *package)
{
    if (getuid() == 0) {
        printf("# file: %s\nbox -i %s\n", path, package);
    } else {
        printf("# file: %s\nsudo box -i %s\n", path, package);
    }
}

static void handle_not_found(GList *list, gchar *file, gboolean new_abi)
{
    GList *li = NULL;
    SearchResult *search_result;
    PackageSearch *search;
    gchar *full_path;

    if (list != NULL)
    {
        printf("\nRequested file '%s' cannot be found, but you can simply install it!\n\n", file);
    }
    if (new_abi)
    {
        for (li = list; li != NULL; li = li->next)
        {
            search = (PackageSearch*)li->data;
            full_path = g_build_filename(dest_dir, search->location, NULL);
            print_install_command(full_path, search->package);
            g_free(full_path);
        }
    } else
    {
        for (li = list; li != NULL; li = li->next)
        {
            search_result = (SearchResult*)li->data;
            full_path = g_build_filename(dest_dir, search_result->file, NULL);
            print_install_command(full_path, search_result->package);
            g_free(full_path);
        }
    }
    if (list != NULL)
    {
        printf("\n");
    }
}

int main (int argc, char *argv[])
{
    GError *error = NULL;
    GOptionContext *context;

    GList *files = NULL;
    sqlite3 *db = NULL;

    context = g_option_context_new ("box-not-found");
    g_option_context_add_main_entries (context, entries, NULL);
    g_option_context_set_help_enabled(context, TRUE);
    g_option_group_set_error_hook(g_option_context_get_main_group(context), (GOptionErrorFunc)goption_error_func);
    g_option_context_parse (context, &argc, &argv, &error);


    db = box_db_open(dest_dir);

    if (!box_db_attach_repos(db, box_get_repos_dir(dest_dir))) {
        printf("No indexes found. Run 'box --sync-repos' first.\n");

        g_option_context_free(context);
        box_db_close(db);
	return 0;
    }
    box_db_repos_init(db);

    if (argc == 2)
    {

        if (strcmp(arg_mode, "file") == 0)
        {
            files = box_db_repos_search_file(db, argv[1]);
            handle_not_found(files, argv[1], TRUE);
            box_db_repos_package_list_free(files);
        } else if (strcmp(arg_mode, "command") == 0)
        {
            files = box_db_repos_search_command(db, argv[1]);
            handle_not_found(files, argv[1], FALSE);
            box_db_search_result_list_free(files);
        } else if (strcmp(arg_mode, "man") == 0)
        {
            files = box_db_repos_search_man(db, argv[1]);
            handle_not_found(files, argv[1], FALSE);
            box_db_search_result_list_free(files);
        } else if (strcmp(arg_mode, "header") == 0)
        {
            files = box_db_repos_search_header(db, argv[1]);
            handle_not_found(files, argv[1], FALSE);
            box_db_search_result_list_free(files);
        } else if (strcmp(arg_mode, "pkgconfig") == 0)
        {
            files = box_db_repos_search_pkgconfig(db, argv[1]);
            handle_not_found(files, argv[1], FALSE);
            box_db_search_result_list_free(files);
        } else {
            printf("Unknown mode.\n");
        }
    } else {
        printf("You must specify what you are looking for.\n");
    }

    g_option_context_free(context);

    box_db_close(db);

    return 0;
}

