#include <glib.h>
#include <glib/gi18n.h>
#include <stdio.h>
#include <string.h>

#include "libbox-db.h"
#include "libbox-db-deps.h"
#include "libbox-db-utils.h"
#include "libbox-parser.h"
#include "libinternal-utils.h"

static gchar* arg_package = NULL;
static gchar* arg_deps = NULL;
static gchar* arg_files = NULL;
static gchar* arg_search = NULL;
static gchar* dest_dir = "/";
static gboolean arg_installed = FALSE;
static gboolean arg_dup = FALSE;

static GOptionEntry entries[] = 
{
    { "package", 'p', 0, G_OPTION_ARG_STRING, &arg_package, "Show information about installed package", NULL },
    { "deps", 'd', 0, G_OPTION_ARG_STRING, &arg_deps, "Show package's dependencies", NULL },
    { "files", 'f', 0, G_OPTION_ARG_STRING, &arg_files, "Show package's files", NULL },
    { "search", 's', 0, G_OPTION_ARG_STRING, &arg_search, "Search for installed files", NULL },
    { "installed", 'i', 0, G_OPTION_ARG_NONE, &arg_installed, "Show installed packages", NULL },
    { "dup", 0, 0, G_OPTION_ARG_NONE, &arg_dup, "Show file duplicates", NULL },
    { "root", 0, 0, G_OPTION_ARG_STRING, &dest_dir, "Specify installation directory (default is '/')", NULL },
    { NULL }
};

static void print_list_str(GList *list, gchar* prefix)
{
    GList *li = NULL;
    DependencyInfo *dep_info = NULL;

    for (li = list; li != NULL; li = li->next)
    {
        dep_info = (DependencyInfo*)li->data;
        if (prefix)
            printf("%s| %-80.80s | %-20.20s | %s\n", prefix, dep_info->name, dep_info->state, dep_info->package);
        else
            printf("%-80.80s | %-20.20s | %s\n", dep_info->name, dep_info->state, dep_info->package);
    }
}

int main (int argc, char *argv[])
{
    GError *error = NULL;
    GOptionContext *context;

    GList *files = NULL;
    GList *li = NULL;
    gchar *full_path = NULL;
    sqlite3 *db = NULL;

    context = g_option_context_new ("box-query");
    g_option_context_add_main_entries (context, entries, NULL);
    g_option_context_set_help_enabled(context, TRUE);
    g_option_group_set_error_hook(g_option_context_get_main_group(context), (GOptionErrorFunc) goption_error_func);
    g_option_context_parse (context, &argc, &argv, &error);

    db = box_db_open(dest_dir);

    if ((arg_files) || (arg_search) || (arg_installed) || (arg_dup) || (arg_deps))
    {
        if (arg_files)
        {
            files = box_db_get_file_list(db, arg_files);
            for (li = files; li != NULL; li = li->next)
            {
                full_path = g_build_filename(dest_dir, (gchar *)li->data, NULL); 
                printf("%s\n", full_path);
                g_free(full_path);
            }
            box_db_file_list_free(files);
        } else if (arg_search)
        {
            SearchResult *search_result = NULL;

            files = box_db_search_file(db, arg_search);
            for (li = files; li != NULL; li = li->next)
            {
                search_result = (SearchResult*)li->data;
                full_path = g_build_filename(dest_dir, search_result->file, NULL); 
                printf("%s\t%s\n", full_path, search_result->package);
                g_free(full_path);
            }
            box_db_search_result_list_free(files);
        } else if (arg_installed)
        {
            PackageInfo *package = NULL;

            files = box_db_get_installed_packages(db);
            for (li = files; li != NULL; li = li->next)
            {
                package = (PackageInfo*)li->data;
                printf("%s\t%s\t%s\t%s\n", package->name, package->version, package->release, package->inst_date);
                box_db_package_info_free(package);
            }
            g_list_free(files);
        } else if (arg_dup)
        {
            SearchResult *search_result = NULL;

            files = box_db_search_dup_files(db);
            for (li = files; li != NULL; li = li->next)
            {
                search_result = (SearchResult*)li->data;
                full_path = g_build_filename(dest_dir, search_result->file, NULL); 
                printf("%s\t%s\n", full_path, search_result->package);
                g_free(full_path);
            }
            box_db_search_result_list_free(files);
        } else if (arg_deps)
        {
            files = box_db_deps_get_requires(db, arg_deps);
            print_list_str(files, "R  ");
            box_db_deps_list_free(files);

            files = box_db_deps_get_build_requires(db, arg_deps);
            print_list_str(files, "BR ");
            box_db_deps_list_free(files);

            files = box_db_deps_get_provides(db, arg_deps);
            print_list_str(files, "P  ");
            box_db_deps_list_free(files);

            files = box_db_deps_get_build_provides(db, arg_deps);
            print_list_str(files, "BP ");
            box_db_deps_list_free(files);

        }
    } else {
        g_print("Wrong parameters.\n");
        goption_show_error(context);
    }

    g_option_context_free(context);


    box_db_close(db);


    return 0;
}
