#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <sqlite3.h>

#include "libbox-db.h"
#include "libbox-db-utils.h"
#include "libbox-db-deps.h"

static void box_db_insert_dep(sqlite3 *db, gint package_id, gchar *depname, gint deps_req_prov, gint deps_type)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    gchar *sql = NULL;
    gint id = -1;
    gchar **lines = NULL;
    gchar *dep_name = NULL;
    gchar *dep_state = NULL;

    lines = g_strsplit(depname, "\t", 0);
    dep_name = lines[0];
    g_assert(depname != NULL);
    if (lines[1])
        dep_state = lines[1];
    else
        dep_state = "";

    sql = g_strdup_printf("SELECT dep_id FROM dep WHERE name = '%s' AND type = %d", dep_name, deps_type); 
    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            id = sqlite3_column_int(stmt, 0);
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);
    g_free(sql);

    if (id == -1)
    {
        sql = g_strdup_printf("INSERT INTO dep (dep_id, name, type) VALUES (NULL, '%s', %d)", dep_name, deps_type);
        box_db_execute_simple(db, sql);
        g_free(sql);
        id = sqlite3_last_insert_rowid(db);
    }


    sql = g_strdup_printf("INSERT INTO dep_list (dep_id, package_id, reqprov, state) VALUES (%d, %d, %d, '%s')", id, package_id, deps_req_prov, dep_state);
    box_db_execute_simple(db, sql);
    g_free(sql);

    g_strfreev(lines);
}

static GList*
box_db_parse_dep_file(const gchar *filename, gboolean split)
{
    gchar *file_buffer = NULL;
    GError *error = NULL;
    gchar **lines = NULL;
    gchar **lines2 = NULL;
    gint i = 0;
    GList *result = NULL;

    if (!g_file_get_contents(filename, &file_buffer, NULL, &error))
    {
        /*g_print("Couldn't read file \"%s\"!\n", filename); */
        return NULL;
    }

    lines = g_strsplit(file_buffer, "\n", 0);
    g_free(file_buffer);

    while (lines[i])
    {
        if (strlen(lines[i]) > 0) {
            if (split)
            {
                lines2 = g_strsplit(lines[i], "\t", 0);
                result = g_list_prepend(result, g_strdup(lines2[0]));
                g_strfreev(lines2);
            } else
            {
                result = g_list_prepend(result, g_strdup(lines[i]));
            }
        }
        i++;
    }

    g_strfreev(lines);

    return result;
}


GList*
box_db_check_deps_upgrade(sqlite3 *db, const gchar *filename, const char *name)
{
    GList *result;
    GList *deps = NULL;
    GList *deps_new, *deps_old, *li1, *li2;
    gchar *sql;
    gboolean found;

    sql = g_strdup_printf("SELECT d.name FROM dep_list AS dl JOIN dep AS d ON d.dep_id=dl.dep_id JOIN package AS p ON p.package_id=dl.package_id WHERE p.name = '%s' AND dl.reqprov = 1 AND d.type = 0", name);

    deps_new = box_db_parse_dep_file(filename, TRUE);
    deps_old = box_db_execute_simple_query_get_string_list(db, sql);
    g_free(sql);

    for (li1 = deps_old; li1 != NULL; li1 = li1->next)
    {
        found = FALSE;
        for (li2 = deps_new; li2 != NULL; li2 = li2->next)
        {
            /*printf("%s\t%s\n", (char*)li1->data, (char *)li2->data); */
            if (strcmp((char*)li1->data, (char *)li2->data) == 0)
                found = TRUE;
        }
        if (!found)
        {
            printf("%s\n", (char*)li1->data);
            deps = g_list_prepend(deps, (gchar*)li1->data);
        }
    }

    result = box_db_repos_what_require_many(db, deps, name);

    /* TODO: free memory */

    return result;
}

static void box_db_parse_deps(sqlite3 *db, gint package_id, const gchar *filename, gint deps_req_prov, gint deps_type)
{
    gchar *sql = NULL;
    GList *deps, *li;

    deps = box_db_parse_dep_file(filename, FALSE);

    if (!deps)
        return;

    /* DELETE old dependencies associations */
    sql = g_strdup_printf("DELETE FROM dep_list WHERE dep_id IN (SELECT l.dep_id FROM dep_list AS l, dep AS d WHERE d.dep_id = l.dep_id AND l.package_id = %d AND d.type = %d AND l.reqprov = %d) AND package_id = %d", package_id, deps_type, deps_req_prov, package_id);
    box_db_execute_simple(db, sql);
    g_free(sql);

    for (li=deps; li != NULL; li = li->next)
    {
        box_db_insert_dep(db, package_id, li->data, deps_req_prov, deps_type);
    }

    box_db_file_list_free(deps);
}

void box_db_install_requires(sqlite3 *db, gint package_id, const gchar *filename)
{
    box_db_parse_deps(db, package_id, filename, DEPS_REQUIRE, DEPS_NORMAL);
}

void box_db_install_build_requires(sqlite3 *db, gint package_id, const gchar *filename)
{
    box_db_parse_deps(db, package_id, filename, DEPS_REQUIRE, DEPS_BUILD);
}

void box_db_install_provides(sqlite3 *db, gint package_id, const gchar *filename)
{
    box_db_parse_deps(db, package_id, filename, DEPS_PROVIDE, DEPS_NORMAL);
}

void box_db_install_build_provides(sqlite3 *db, gint package_id, const gchar *filename)
{
    box_db_parse_deps(db, package_id, filename, DEPS_PROVIDE, DEPS_BUILD);
}



static GList* box_db_get_deps(sqlite3 *db, const gchar *package, gint deps_req_prov, gint deps_type)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *deps = NULL;
    gchar *sql = NULL;
    DependencyInfo *dep_info = NULL;
    gchar *pkg = NULL, *pkg_tmp = NULL;

    sql = g_strdup_printf("SELECT d.name, r.state, (SELECT pp.name FROM dep AS dd, dep_list AS rr, package AS pp WHERE dd.dep_id = rr.dep_id AND rr.package_id = pp.package_id AND rr.reqprov = 1 AND dd.name = d.name ORDER BY dd.name LIMIT 1) AS pkg FROM dep AS d, dep_list AS r, package AS p WHERE d.dep_id = r.dep_id AND r.package_id = p.package_id AND p.name = '%s' AND d.type = %d AND r.reqprov = %d ORDER BY pkg DESC, d.name", package, deps_type, deps_req_prov);
    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);
    
    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            dep_info = g_new0(DependencyInfo, 1);
            dep_info->name = g_strdup((gchar *)sqlite3_column_text(stmt, 0));
            dep_info->state = g_strdup((gchar *)sqlite3_column_text(stmt, 1));
            pkg_tmp = (gchar *)sqlite3_column_text(stmt, 2);
            if (pkg_tmp == NULL)
                pkg = g_strdup("?");
            else
                pkg = g_strdup(pkg_tmp);
            dep_info->package = pkg;
            deps = g_list_prepend(deps, dep_info);
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);

    g_free(sql);

    return deps;
}

GList* box_db_deps_get_requires(sqlite3 *db, const gchar *package)
{
    return box_db_get_deps(db, package, DEPS_REQUIRE, DEPS_NORMAL);
}

GList* box_db_deps_get_build_requires(sqlite3 *db, const gchar *package)
{
    return box_db_get_deps(db, package, DEPS_REQUIRE, DEPS_BUILD);
}

GList* box_db_deps_get_provides(sqlite3 *db, const gchar *package)
{
    return box_db_get_deps(db, package, DEPS_PROVIDE, DEPS_NORMAL);
}

GList* box_db_deps_get_build_provides(sqlite3 *db, const gchar *package)
{
    return box_db_get_deps(db, package, DEPS_PROVIDE, DEPS_BUILD);
}

void box_db_deps_free(DependencyInfo *dep_info)
{
    g_free(dep_info->name);
    g_free(dep_info->state);
    g_free(dep_info->package);
    g_free(dep_info);
}

void box_db_deps_list_free(GList *list)
{
    GList *li;

    for (li = list; li != NULL; li = li->next)
    {
        box_db_deps_free((DependencyInfo*)li->data);
    }
    g_list_free(list);
}


GList* box_db_remove_deps(sqlite3 *db, const gchar *package, gboolean force)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *list = NULL;
    gchar *sql = NULL;

    sql = g_strdup_printf("SELECT DISTINCT pp.name FROM dep AS dd, dep_list AS ddl, package AS pp WHERE dd.dep_id = ddl.dep_id AND ddl.reqprov = 0 AND ddl.package_id = pp.package_id AND dd.name IN (SELECT d.name FROM dep AS d, dep_list AS dl, package AS p WHERE d.dep_id = dl.dep_id AND dl.reqprov = 1 AND d.type = 0 AND dl.package_id = p.package_id AND p.name = '%s') AND pp.name != '%s' ORDER BY pp.name DESC", package, package);
    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);
    
    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            list = g_list_prepend(list, g_strdup((gchar *)sqlite3_column_text(stmt, 0)));
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);
    g_free(sql);

    /* if some of packages depends on the package then don't remove dependency */
    /* and return list of packages which depends on */
    if (list != NULL && !force)
        return list;

    sql = g_strdup_printf("DELETE FROM dep WHERE name IN (SELECT d.name FROM dep AS d, dep_list AS l WHERE d.dep_id = l.dep_id GROUP BY d.name HAVING COUNT(l.package_id) = 1 AND l.package_id = (SELECT package_id FROM package WHERE name = '%s'))", package);
    box_db_execute_simple(db, sql);
    g_free(sql);

    sql = g_strdup_printf("DELETE FROM dep_list WHERE package_id = (SELECT package_id FROM package WHERE name = '%s')", package);
    box_db_execute_simple(db, sql);
    g_free(sql);

    return NULL;
}

