#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sqlite3.h>
#include <glib/gstdio.h>

#include "libbox.h"
#include "libbox-db.h"
#include "libbox-db-repos.h"
#include "libbox-db-utils.h"
#include "libbox-repos.h"

#define DEFAULT_ARCH "i686"
#define GENERATE_GRAPH 0
#define USE_TSORT 1

#ifdef __HAIKU__
#define HAIKU_SQL
#endif
#define HAIKU_SQL

/*
 * TODO:
 * - add GUI filtering for available and all packages
 * - add filtering for box_db_repos_search_file()
 */

GList *REPOS = NULL;

gboolean box_db_attach_repo(sqlite3 *db, const gchar *destdir, const gchar *repo_name)
{
    gchar *sql = NULL;
    gchar *dbdir;
    gchar *file;

    dbdir = box_get_db_dir();

    file = g_strdup_printf("%s%s/index-%s", destdir, dbdir, repo_name);
    if (!g_file_test(file, G_FILE_TEST_EXISTS)) {
        g_free(file);
        g_free(dbdir);
        return FALSE;
    }
    sql = g_strdup_printf("ATTACH DATABASE '%s' AS '%s'", file, repo_name);
    g_free(file);
    g_free(dbdir);
    box_db_execute_simple(db, sql);
    g_free(sql);

    REPOS = g_list_append(REPOS, g_strdup(repo_name));

    return TRUE;
}

gboolean box_db_attach_repos(sqlite3 *db, const gchar *destdir)
{
    GList *repos, *li;
    RepoInfo *repo;
    gboolean ret = TRUE;

    repos = box_repos_list_get();

    for (li = repos; li != NULL; li=li->next)
    {
        repo = (RepoInfo*) li->data;
        if (repo->enabled)
        {
            if (!box_db_attach_repo(db, destdir, repo->name))
                ret = FALSE;
        }
    }

    box_repos_list_free(repos);

    return ret;
}

static gint compare_string_func(gconstpointer str1, gconstpointer str2)
{
    return strcmp(str1, str2);
}

void box_db_detach_repo(sqlite3 *db, const gchar *repo_name)
{
    gchar *sql = NULL;
    GList *li = NULL;

    sql = g_strdup_printf("DETACH DATABASE '%s'", repo_name);
    box_db_execute_simple(db, sql);

    li = g_list_find_custom( REPOS, repo_name, compare_string_func);
    g_assert(li != NULL);

    REPOS = g_list_remove(REPOS, li->data);

    g_free(sql);
}

void box_db_detach_repos(sqlite3 *db)
{
    gchar *sql = NULL;
    GList *li = NULL;
    gchar *repo_name;

    for (li = REPOS; li != NULL; li = li->next)
    {
        repo_name = (gchar*) li->data;
        sql = g_strdup_printf("DETACH DATABASE '%s'", repo_name);
        box_db_execute_simple(db, sql);
        g_free(sql);
        g_free(repo_name);
    }

    g_list_free(REPOS);
    REPOS = NULL;
}


GList* box_db_get_repos()
{
    return REPOS;
}

void box_db_repos_init(sqlite3 *db)
{
    GList *li = NULL;
    gchar *tmp = NULL;
    gchar *sql = NULL;
    gchar **str_array = NULL;
    gint idx = 0;

    str_array = g_new(gchar*, g_list_length(REPOS) + 1);
    for (li = REPOS; li != NULL; li = li->next)
    {
        str_array[idx++] = g_strdup_printf(""
"            SELECT *,"
"                (SELECT value"
"                 FROM %s.meta"
"                 WHERE name = 'location') AS location, '%s' AS reponame"
"            FROM %s.package",
            (gchar *)li->data, (gchar *)li->data, (gchar *)li->data);
    }
    str_array[idx++] = NULL;

    if (g_list_length(REPOS) > 0)
    {
        tmp = g_strjoinv(" UNION ", str_array);
        sql = g_strdup_printf("DROP VIEW IF EXISTS u_package; CREATE TEMPORARY VIEW u_package AS %s", tmp);
        g_free(tmp);

        box_db_execute_simple(db, sql);
        g_free(sql);
    }

    g_strfreev(str_array);

    /* Create temporary views used for dependencies resolving */
    for (li = REPOS; li != NULL; li = li->next)
    {
        sql = g_strdup_printf(""
"            DROP TABLE IF EXISTS provide_%s;"
"            CREATE TEMPORARY TABLE provide_%s AS"
"                SELECT p.name AS package, p.version AS version, p.summary AS summary, d.name AS dep"
"                FROM %s.dep AS d"
"                JOIN %s.dep_list AS dl ON d.dep_id = dl.dep_id"
"                JOIN %s.package AS p ON dl.package_id = p.package_id"
"                WHERE d.type = 0 AND dl.reqprov = 1",
            (gchar *)li->data, (gchar *)li->data, (gchar *)li->data, (gchar *)li->data, (gchar *)li->data);

        box_db_execute_simple(db, sql);
        g_free(sql);

        sql = g_strdup_printf(""
"            DROP TABLE IF EXISTS require_%s;"
"            CREATE TEMPORARY TABLE IF NOT EXISTS require_%s AS"
"                SELECT p.name AS package, p.version AS version, p.summary AS summary, d.name AS dep"
"                FROM %s.dep AS d"
"                JOIN %s.dep_list AS dl ON d.dep_id = dl.dep_id"
"                JOIN %s.package AS p ON dl.package_id = p.package_id"
"                WHERE d.type = 0 AND dl.reqprov = 0",
            (gchar *)li->data, (gchar *)li->data, (gchar *)li->data, (gchar *)li->data, (gchar *)li->data);

        box_db_execute_simple(db, sql);
        g_free(sql);
    }
}

static gchar* get_master_package(gchar *package)
{
    /* FIXME: small memleaks are here? */
    if (g_str_has_suffix(package, "-devel"))
        package[strlen(package) - 6] = 0;
    if (g_str_has_suffix(package, "-static"))
        package[strlen(package) - 7] = 0;

    return package;
}

static PackageSearch*
box_db_repos_package_search_new(gchar *package, gchar *version, gchar *description, gchar *arch,
        gboolean installed, gchar *location, gchar *reponame, gboolean gui)
{
    PackageSearch *search = g_new0(PackageSearch, 1);

    search->package = package;
    search->version = version;
    if (description == NULL) {
        search->description = g_strdup("");
    } else {
        search->description = description;
    }
    search->arch = arch;
    search->installed = installed;
    search->location = location;
    search->reponame = reponame;
    search->gui = gui;

    return search;
}

static void box_db_repos_package_search_free(PackageSearch *search)
{
    g_free(search->package);
    g_free(search->version);
    g_free(search->description);
    g_free(search->arch);
    g_free(search->location);
    g_free(search->reponame);
    g_free(search);
}

void box_db_repos_package_list_free(GList *list)
{
    GList *li;

    for (li = list; li != NULL; li = li->next)
    {
        box_db_repos_package_search_free((PackageSearch*)li->data);
    }
    g_list_free(list);
}


GList* box_db_repos_packages_for_upgrade(sqlite3 *db)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *files = NULL;
    gchar *sql = NULL;
    gchar *dirname = NULL;
    PackageSearch *search = NULL;

    sql = g_strdup_printf(""
"        SELECT p1.location, p1.name, p1.version, p1.`release`, p1.summary, p1.version, p1.reponame"
"        FROM package AS p0"
"        JOIN u_package AS p1 ON p0.name = p1.name"
"        WHERE p0.`release` < p1.`release`");

    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            dirname = g_strdup((gchar *)sqlite3_column_text(stmt, 1));
            dirname = get_master_package(dirname);
            search = box_db_repos_package_search_new(
                    g_strdup((gchar *)sqlite3_column_text(stmt, 1)),
                    g_strdup((gchar *)sqlite3_column_text(stmt, 5)),
                    g_strdup((gchar *)sqlite3_column_text(stmt, 4)),
                    g_strdup(DEFAULT_ARCH),
                    TRUE,
                    g_strdup_printf("%s%s/%s/%s-%s-%s.i686.box", 
                        (gchar *)sqlite3_column_text(stmt, 0),
                        dirname,
                        (gchar *)sqlite3_column_text(stmt, 3),
                        (gchar *)sqlite3_column_text(stmt, 1),
                        (gchar *)sqlite3_column_text(stmt, 2),
                        (gchar *)sqlite3_column_text(stmt, 3)),
                    g_strdup((gchar *)sqlite3_column_text(stmt, 6)),
                    FALSE /* FIXME minor: GUI or not? */
            );
            files = g_list_prepend(files, search);
            g_free(dirname);
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);

    g_free(sql);

    return files;
}

static gchar* box_db_repos_get_filter(gchar *field, gchar *field_desc, const gchar *package, gint filter_flags)
{
    gchar *filter = NULL;
    gchar *gui_filter = NULL;
    gchar *desc_filter = NULL;
    gboolean details = FALSE;

    if (((filter_flags & PKG_GUI) == PKG_GUI) && ((filter_flags & PKG_TEXT) == PKG_TEXT)) {
        /* do nothing */
    } else if ((filter_flags & PKG_GUI) == PKG_GUI) {
        gui_filter = g_strdup(" AND gui = 1");
    } else if ((filter_flags & PKG_TEXT) == PKG_TEXT) {
        gui_filter = g_strdup(" AND gui = 0");
    }
    if (gui_filter == NULL) {
        gui_filter = g_strdup("");
    }

    details = ((filter_flags & PKG_SEARCH_DETAILS) == PKG_SEARCH_DETAILS);
    if (!details)
        desc_filter = g_strdup("");

    if (filter_flags == PKG_NONE || (((filter_flags & PKG_DEVEL) == PKG_DEVEL) && ((filter_flags & PKG_NON_DEVEL) == PKG_NON_DEVEL)))
    {
        if (details)
            desc_filter = g_strdup_printf("OR (%s LIKE '%%%s%%')", field_desc, package);

        filter = g_strdup_printf("((%s LIKE '%%%s%%') %s )%s", field, package, desc_filter, gui_filter);
    } else if ((filter_flags & PKG_DEVEL) == PKG_DEVEL)
    {
        if (details)
            desc_filter = g_strdup_printf("OR (%s LIKE '%%%s%%-devel')", field_desc, package);
        filter = g_strdup_printf("((%s LIKE '%%%s%%-devel') %s )%s", field, package, desc_filter, gui_filter);
    } else if ((filter_flags & PKG_NON_DEVEL) == PKG_NON_DEVEL)
    {
        if (details)
            desc_filter = g_strdup_printf("OR (%s LIKE '%%%s%%' AND %s NOT LIKE '%%-devel')", field_desc, package, field_desc);
        filter = g_strdup_printf("((%s LIKE '%%%s%%' AND %s NOT LIKE '%%-devel') %s )%s", field, package, field, desc_filter, gui_filter);
    }

    g_free(desc_filter);
    g_free(gui_filter);

    return filter;
}


GList* box_db_repos_packages_search_by_data(sqlite3 *db, const gchar *package_name, const gchar *package_version)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *files = NULL;
    gchar *sql = NULL;
    PackageSearch *search = NULL;

    sql = g_strdup_printf(""
"        SELECT up.name, CASE WHEN p.package_id IS NULL THEN 0 ELSE 1 END, up.summary, up.version, up.reponame"
"        FROM u_package AS up"
"        LEFT OUTER JOIN package AS p ON up.name = p.name"
"        WHERE up.name = '%s' AND up.version = '%s'"
"        LIMIT 1", package_name, package_version);

    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            search = box_db_repos_package_search_new(
                        g_strdup((gchar *)sqlite3_column_text(stmt, 0)), 
                        g_strdup((gchar *)sqlite3_column_text(stmt, 3)),
                        g_strdup((gchar *)sqlite3_column_text(stmt, 2)),
                        g_strdup(DEFAULT_ARCH),
                        sqlite3_column_int(stmt, 1),
                        g_strdup(""),
                        g_strdup((gchar *)sqlite3_column_text(stmt, 4)),
                        FALSE /* FIXME minor: GUI or not? */
                    );
            files = g_list_prepend(files, search);
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);

    g_free(sql);

    return files;
}


GList* box_db_repos_packages_search_installed(sqlite3 *db, const gchar *package, gint filter_flags)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *files = NULL;
    gchar *sql = NULL;
    PackageSearch *search = NULL;
    gchar *filter;

    filter = box_db_repos_get_filter("name", "summary", package, filter_flags);
    sql = g_strdup_printf("SELECT name, summary, version,"
"                (SELECT COUNT(*)"
"                FROM dep_list AS xdl, dep AS xd"
"                WHERE xdl.package_id = p.package_id AND"
"                    xdl.reqprov = 0 AND"
"                    xd.dep_id = xdl.dep_id AND"
"                    xd.name = 'soname.libgtk-x11-2.0.so.0'"
"            ) AS gui"
"            FROM package AS p"
"            WHERE %s"
"            ORDER BY name DESC", filter);
    g_free(filter);

    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            search = box_db_repos_package_search_new(
                        g_strdup((gchar *)sqlite3_column_text(stmt, 0)), 
                        g_strdup((gchar *)sqlite3_column_text(stmt, 2)),
                        g_strdup((gchar *)sqlite3_column_text(stmt, 1)),
                        g_strdup(DEFAULT_ARCH),
                        TRUE,
                        g_strdup(""),
                        g_strdup("installed"),
                        sqlite3_column_int(stmt, 3)
                    );
            files = g_list_prepend(files, search);
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);

    g_free(sql);

    return files;
}


GList* box_db_repos_packages_search_available(sqlite3 *db, const gchar *package, gint filter_flags)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *files = NULL;
    gchar *sql = NULL;
    PackageSearch *search = NULL;
    gchar *filter;

    filter = box_db_repos_get_filter("name", "summary", package, filter_flags);
    /* FIXME major: GUI or not? */
    sql = g_strdup_printf(""
"        SELECT name, summary, version, reponame, 1 AS gui"
"        FROM u_package"
"        WHERE %s AND name NOT IN (SELECT name FROM package)"
"        ORDER BY name DESC", filter);
    g_free(filter);
    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            search = box_db_repos_package_search_new(
                        g_strdup((gchar *)sqlite3_column_text(stmt, 0)),
                        g_strdup((gchar *)sqlite3_column_text(stmt, 2)),
                        g_strdup((gchar *)sqlite3_column_text(stmt, 1)),
                        g_strdup(DEFAULT_ARCH),
                        FALSE,
                        g_strdup(""),
                        g_strdup((gchar *)sqlite3_column_text(stmt, 3)),
                        sqlite3_column_int(stmt, 4)
                    );
            files = g_list_prepend(files, search);
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);

    g_free(sql);

    return files;
}


GList* box_db_repos_packages_search_all(sqlite3 *db, const gchar *package, gint filter_flags)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *files = NULL;
    gchar *sql = NULL;
    PackageSearch *search = NULL;
    gchar *filter = NULL;

    filter = box_db_repos_get_filter("up.name", "up.summary", package, filter_flags);
    /* FIXME major: GUI or not? */
    sql = g_strdup_printf("SELECT"
"                up.name,"
"                CASE WHEN p.package_id IS NULL THEN 0 ELSE 1 END,"
"                up.summary,"
"                up.version,"
"                up.reponame,"
"                1 AS gui"
"            FROM u_package AS up"
"            LEFT OUTER JOIN package AS p"
"            ON up.name = p.name"
"            WHERE %s"
"            ORDER BY up.name DESC", filter);
    g_free(filter);

    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            search = box_db_repos_package_search_new(
                        g_strdup((gchar *)sqlite3_column_text(stmt, 0)), 
                        g_strdup((gchar *)sqlite3_column_text(stmt, 3)),
                        g_strdup((gchar *)sqlite3_column_text(stmt, 2)),
                        g_strdup(DEFAULT_ARCH),
                        sqlite3_column_int(stmt, 1), g_strdup(""),
                        g_strdup((gchar *)sqlite3_column_text(stmt, 4)),
                        sqlite3_column_int(stmt, 5)
                    );
            files = g_list_prepend(files, search);
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);

    g_free(sql);

    return files;
}

GList* box_db_repos_packages_search_one(sqlite3 *db, const gchar *package)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *files = NULL;
    gchar *sql = NULL;
    PackageSearch *search = NULL;

    /* FIXME minor: GUI or not? */
    sql = g_strdup_printf(""
"        SELECT name, summary, version, reponame, 1 AS gui"
"        FROM u_package"
"        WHERE name = '%s'"
"        LIMIT 1", package);

    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            search = box_db_repos_package_search_new(
                        g_strdup((gchar *)sqlite3_column_text(stmt, 0)),
                        g_strdup((gchar *)sqlite3_column_text(stmt, 2)),
                        g_strdup((gchar *)sqlite3_column_text(stmt, 1)),
                        g_strdup(DEFAULT_ARCH),
                        FALSE, /* FIXME minor: installed or not? */
                        g_strdup(""),
                        g_strdup((gchar *)sqlite3_column_text(stmt, 3)),
                        sqlite3_column_int(stmt, 4)
                    );
            files = g_list_prepend(files, search);
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);

    g_free(sql);

    return files;
}


GList* box_db_repos_package_get_for_installation(sqlite3 *db, const gchar *package)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *files = NULL;
    gchar *sql = NULL;
    gchar *dirname = NULL;
    gchar *file = NULL;
    PackageSearch *search = NULL;

    sql = g_strdup_printf(""
"        SELECT p1.location, p1.name, p1.version, p1.`release`, p1.summary, p1.reponame"
"        FROM u_package AS p1"
"        WHERE p1.name = '%s'", package);
    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            dirname = g_strdup((gchar *)sqlite3_column_text(stmt, 1));
            dirname = get_master_package(dirname);
            file = g_strdup_printf("%s-%s-%s.i686.box", (gchar *)sqlite3_column_text(stmt, 1), (gchar *)sqlite3_column_text(stmt, 2), (gchar *)sqlite3_column_text(stmt, 3));
            search = box_db_repos_package_search_new(
                        g_strdup((gchar *)sqlite3_column_text(stmt, 1)),
                        g_strdup((gchar *)sqlite3_column_text(stmt, 2)),
                        g_strdup((gchar *)sqlite3_column_text(stmt, 4)),
                        g_strdup(DEFAULT_ARCH),
                        0,
                        g_build_filename((gchar *)sqlite3_column_text(stmt, 0), dirname, (gchar *)sqlite3_column_text(stmt, 3), file, NULL ),
                        g_strdup((gchar *)sqlite3_column_text(stmt, 5)),
                        FALSE
                    );
            g_free(file);
            files = g_list_prepend(files, search);
            g_free(dirname);
            
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);

    g_free(sql);

    return files;
}

static GList*
_box_db_repos_what_provide_require(sqlite3 *db, const gchar *dep, gint type, const gchar *skip_name)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *li = NULL;
    GList *packages = NULL;
    gchar *tmp = NULL;
    gchar *sql = NULL;
    gchar **str_array = NULL;
    gint idx = 0;
    PackageSearch *search = NULL;

    str_array = g_new(gchar*, g_list_length(REPOS) + 1);
    for (li = REPOS; li != NULL; li = li->next)
    {
        str_array[idx++] = g_strdup_printf(""
"            SELECT p.name AS name, p.version, p.summary"
"            FROM %s.package AS p, %s.dep AS d, %s.dep_list AS dl"
"            WHERE d.dep_id = dl.dep_id AND dl.package_id = p.package_id AND dl.reqprov = %d AND (%s) AND p.name != '%s'",
            (gchar *)li->data, (gchar *)li->data, (gchar *)li->data, type, dep, skip_name);
    }
    str_array[idx++] = NULL;

    tmp = g_strjoinv(" UNION ", str_array);
    sql = g_strdup_printf("SELECT * FROM (%s)", tmp);
    g_free(tmp);
    
    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            search = box_db_repos_package_search_new(
                        g_strdup((gchar *)sqlite3_column_text(stmt, 0)),
                        g_strdup((gchar *)sqlite3_column_text(stmt, 1)),
                        g_strdup((gchar *)sqlite3_column_text(stmt, 2)),
                        g_strdup(DEFAULT_ARCH),
                        0,
                        g_strdup(""),
                        g_strdup(""),
                        FALSE
                    );
            packages = g_list_prepend(packages, search);
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);

    g_strfreev(str_array);
    g_free(sql);

    return packages;
}

GList* box_db_repos_what_provide(sqlite3 *db, const gchar *dep)
{
    gchar *str;
    GList *result;

    str = g_strdup_printf("d.name = '%s'", dep);
    result = _box_db_repos_what_provide_require(db, str, 1, "");
    g_free(str);

    return result;
}

GList* box_db_repos_what_require(sqlite3 *db, const gchar *dep)
{
    gchar *str;
    GList *result;

    str = g_strdup_printf("d.name = '%s'", dep);
    result = _box_db_repos_what_provide_require(db, str, 0, "");
    g_free(str);

    return result;
}

GList*
box_db_repos_what_require_many(sqlite3 *db, GList *list, const gchar *skip_name)
{
    gchar *str;
    GList *result;

    str = box_glist_join_printf(list, " OR ", "d.name = '%s'");
    result = _box_db_repos_what_provide_require(db, str, 0, skip_name);
    g_free(str);

    return result;
}

GList* box_db_repos_search_file(sqlite3 *db, const gchar *name)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *files = NULL;
    gchar *sql = NULL;
    gint file_id;
    PackageSearch *search = NULL;
    GList *li = NULL;
    gchar *prefix = NULL;

    for (li = box_db_get_repos(); li != NULL; li = li->next)
    {
        prefix = (gchar *)li->data;
        sql = g_strdup_printf(""
"            SELECT file_id, p.name, p.version, p.summary, (SELECT COUNT(*) FROM package WHERE name = p.name)"
"            FROM %s.file_list f, %s.package p"
"            WHERE p.package_id = f.package_id AND file_id IN ("
"                SELECT file_id"
"                FROM %s.file"
"                WHERE name = '%s' AND type != 'D' )"
"            ORDER BY p.name DESC", prefix, prefix, prefix, name);
        rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

        while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
        {
            if (rc == SQLITE_ROW)
            {
                file_id = sqlite3_column_int(stmt, 0);
                search = box_db_repos_package_search_new(
                            g_strdup((gchar *)sqlite3_column_text(stmt, 1)), 
                            g_strdup((gchar *)sqlite3_column_text(stmt, 2)),
                            g_strdup((gchar *)sqlite3_column_text(stmt, 3)),
                            g_strdup(DEFAULT_ARCH),
                            sqlite3_column_int(stmt, 4),
                            box_db_get_file_path_from_id(db, file_id, prefix),
                            g_strdup(prefix),
                            FALSE
                       );
                files = g_list_prepend(files, search);
            } else if (rc != SQLITE_OK)
            {
                printf("ERROR: %s\n", sqlite3_errmsg(db));
            }
        }
        sqlite3_finalize(stmt);
        g_free(sql);
    }

    return files;
}
GList* box_db_repos_search_file_with_filter(sqlite3 *db, const gchar *name, gint filter_flags)
{
    /* TODO: implement me */
    return box_db_repos_search_file(db, name);
}

GList* box_db_repos_search_man(sqlite3 *db, const gchar *name)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *files = NULL;
    gchar *sql = NULL;
    gint file_id;
    SearchResult *search_result = NULL;
    GList *li = NULL;
    gchar *prefix = NULL;

    for (li = box_db_get_repos(); li != NULL; li = li->next)
    {
        prefix = (gchar *)li->data;
        sql = g_strdup_printf(""
"            SELECT f.file_id, p.name"
"            FROM %s.file AS f"
"            JOIN %s.file_list AS fl ON fl.file_id = f.file_id"
"            JOIN %s.package AS p ON p.package_id = fl.package_id"
"            WHERE parent_id IN ("
"                SELECT file_id"
"                FROM %s.file"
"                WHERE name LIKE 'man%%' AND parent_id = ("
"                    SELECT file_id"
"                    FROM %s.file"
"                    WHERE name = 'man' AND parent_id = ("
"                        SELECT file_id"
"                        FROM %s.file"
"                        WHERE name = 'share' AND parent_id = ("
"                            SELECT file_id"
"                            FROM file"
#ifdef HAIKU_SQL
"                            WHERE name = 'common' AND parent_id = ("
"                                SELECT file_id"
"                                FROM %s.file"
"                                WHERE name = 'boot' AND parent_id = ("
#else
"                                WHERE name = 'usr' AND parent_id = ("
#endif
"                                    SELECT file_id"
"                                    FROM %s.file"
"                                    WHERE name = '' AND parent_id = -1"
"                                    )"
"                                )"
#ifdef HAIKU_SQL
"                            )"
#endif
"                        )"
"                    )"
"                )"
"            AND (f.name LIKE '%s._.gz' OR f.name LIKE '%s.__.gz')",
            prefix, prefix, prefix, prefix, prefix, prefix,
#ifdef HAIKU_SQL
            prefix,
#endif
            prefix, name, name);

        rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

        while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
        {
            if (rc == SQLITE_ROW)
            {
                file_id = sqlite3_column_int(stmt, 0);
                search_result = g_new0(SearchResult, 1);
                search_result->file = box_db_get_file_path_from_id(db, file_id, prefix);
                search_result->package = g_strdup((gchar *)sqlite3_column_text(stmt, 1));
                files = g_list_prepend(files, search_result);
            } else if (rc != SQLITE_OK)
            {
                printf("ERROR: %s\n", sqlite3_errmsg(db));
            }
        }
        sqlite3_finalize(stmt);
        g_free(sql);
    }

    return files;
}

GList* box_db_repos_search_command(sqlite3 *db, const gchar *name)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *files = NULL;
    gchar *sql = NULL;
    gint file_id;
    SearchResult *search_result = NULL;
    GList *li = NULL;
    gchar *prefix = NULL;

    for (li = box_db_get_repos(); li != NULL; li = li->next)
    {
        prefix = (gchar *)li->data;
        sql = g_strdup_printf(""
"        SELECT f.file_id, p.name"
"        FROM %s.file AS f"
"        JOIN %s.file_list AS fl ON fl.file_id = f.file_id"
"        JOIN %s.package AS p ON p.package_id = fl.package_id"
"        WHERE parent_id IN ("
"            SELECT file_id"
"            FROM %s.file"
"            WHERE (name = 'bin' OR name = 'sbin') AND parent_id = ("
"                SELECT file_id"
"                FROM %s.file"
"                WHERE name = '' AND parent_id = -1 )"
"            UNION"
"            SELECT file_id"
"            FROM %s.file"
"            WHERE (name = 'bin' OR name = 'sbin') AND parent_id = ("
"                SELECT file_id"
"                FROM %s.file"
#ifdef HAIKU_SQL
"                WHERE name = 'common' AND parent_id = ("
"                    SELECT file_id"
"                    FROM %s.file"
"                    WHERE name = 'boot' AND parent_id = ("
#else
"                    WHERE name = 'usr' AND parent_id = ("
#endif
"                        SELECT file_id"
"                        FROM %s.file"
"                        WHERE name = '' AND parent_id = -1"
"                        )"
"                    )"
#ifdef HAIKU_SQL
"                )"
#endif
"            )"
"            AND f.name LIKE '%s'",
        prefix, prefix, prefix, prefix, prefix, prefix, prefix,
#ifdef HAIKU_SQL
        prefix,
#endif
        prefix, name);

        rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

        while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
        {
            if (rc == SQLITE_ROW)
            {
                file_id = sqlite3_column_int(stmt, 0);
                search_result = g_new0(SearchResult, 1);
                search_result->file = box_db_get_file_path_from_id(db, file_id, prefix);
                search_result->package = g_strdup((gchar *)sqlite3_column_text(stmt, 1));
                files = g_list_prepend(files, search_result);
            } else if (rc != SQLITE_OK)
            {
                printf("ERROR: %s\n", sqlite3_errmsg(db));
            }
        }
        sqlite3_finalize(stmt);
        g_free(sql);
    }

    return files;
}

GList* box_db_repos_search_header(sqlite3 *db, const gchar *name)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *files = NULL;
    gchar *sql = NULL;
    gint file_id;
    SearchResult *search_result = NULL;
    GList *li = NULL;
    gchar *prefix = NULL;
    gchar *file = NULL;
    gchar *name_file = NULL;
    gchar **tokens = NULL;

    tokens = g_strsplit(name, G_DIR_SEPARATOR_S, 0);
    name_file = g_strdup(tokens[g_strv_length(tokens) - 1]);
    g_strfreev(tokens);

    for (li = box_db_get_repos(); li != NULL; li = li->next)
    {
        prefix = (gchar *)li->data;
        sql = g_strdup_printf(""
"            SELECT f.file_id, p.name"
"            FROM %s.file AS f, %s.file_list AS fl, %s.package AS p"
"            WHERE f.file_id = fl.file_id AND fl.package_id = p.package_id AND f.name LIKE '%s'", prefix, prefix, prefix, name_file);

        rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

        while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
        {
            if (rc == SQLITE_ROW)
            {
                file_id = sqlite3_column_int(stmt, 0);
                file = box_db_get_file_path_from_id(db, file_id, prefix);

                if (g_str_has_prefix(file, PACKAGE_PREFIX"/include/") && g_str_has_suffix(file, name))
                {
                    search_result = g_new0(SearchResult, 1);
                    search_result->file = file;
                    search_result->package = g_strdup((gchar *)sqlite3_column_text(stmt, 1));
                    files = g_list_prepend(files, search_result);
                } else 
                {
                    g_free(file);
                }
            } else if (rc != SQLITE_OK)
            {
                printf("ERROR: %s\n", sqlite3_errmsg(db));
            }
        }
        sqlite3_finalize(stmt);
        g_free(sql);
    }
    g_free(name_file);

    return files;
}

GList* box_db_repos_search_pkgconfig(sqlite3 *db, const gchar *name)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *files = NULL;
    gchar *sql = NULL;
    gint file_id;
    SearchResult *search_result = NULL;
    GList *li = NULL;
    gchar *prefix = NULL;
    gchar *file = NULL;

    for (li = box_db_get_repos(); li != NULL; li = li->next)
    {
        prefix = (gchar *)li->data;
        sql = g_strdup_printf(""
"            SELECT f.file_id, p.name"
"            FROM %s.file AS f, %s.file_list AS fl, %s.package AS p"
"            WHERE f.file_id = fl.file_id AND fl.package_id = p.package_id AND f.name LIKE '%s.pc'", prefix, prefix, prefix, name);

        rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

        while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
        {
            if (rc == SQLITE_ROW)
            {
                file_id = sqlite3_column_int(stmt, 0);
                file = box_db_get_file_path_from_id(db, file_id, prefix);
                if (g_str_has_prefix(file, PACKAGE_LIBDIR"/pkgconfig/"))
                {
                    search_result = g_new0(SearchResult, 1);
                    search_result->file = file;
                    search_result->package = g_strdup((gchar *)sqlite3_column_text(stmt, 1));
                    files = g_list_prepend(files, search_result);
                } else 
                {
                    g_free(file);
                }
            } else if (rc != SQLITE_OK)
            {
                printf("ERROR: %s\n", sqlite3_errmsg(db));
            }
        }
        sqlite3_finalize(stmt);
        g_free(sql);
    }

    return files;
}


gchar* box_db_repos_get_files_string(sqlite3 *db, const gchar *name, const gchar *version)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    gchar *sql = NULL;
    gint file_id;
    GList *li = NULL;
    gchar *prefix = NULL;
    gchar *result = g_strdup("");
    gchar *old_result = NULL;
    gchar *file = NULL;

    for (li = box_db_get_repos(); li != NULL; li = li->next)
    {
        prefix = (gchar *)li->data;
        sql = g_strdup_printf(""
"            SELECT file_id"
"            FROM %s.file_list f, %s.package p"
"            WHERE p.package_id = f.package_id AND p.name = '%s' AND p.version = '%s'", prefix, prefix,  name, version);
        rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

        while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
        {
            if (rc == SQLITE_ROW)
            {
                file_id = sqlite3_column_int(stmt, 0);
                file = box_db_get_file_path_from_id(db, file_id, prefix);
                old_result = result;
                result = g_strconcat(result, file, "\n", NULL);
                g_free(old_result);
                g_free(file);
            } else if (rc != SQLITE_OK)
            {
                printf("ERROR: %s\n", sqlite3_errmsg(db));
            }
        }
        sqlite3_finalize(stmt);
        g_free(sql);
    }

    return result;
}

static gchar*
_get_depends_requires_many_sql(GList *names, const gchar *prefix, gboolean depends)
{
    gchar *str = NULL;
    gchar *str2 = NULL;
    gchar *result;

    str = box_glist_join_printf(names, " OR ", "p.name = '%s'");
    str2 = box_glist_join_printf(names, " AND ", "p2.name != '%s'");

    result = g_strdup_printf(""
"        SELECT p2.name, p2.version, p2.summary"
"        FROM %s.dep AS d, %s.dep_list AS dl, %s.package AS p, %s.dep AS d2, %s.dep_list AS dl2, %s.package AS p2"
"        WHERE (d.dep_id = dl.dep_id AND dl.package_id = p.package_id AND (%s) AND d.type = 0 AND dl.reqprov = %d)"
"              AND (d2.dep_id = dl2.dep_id AND dl2.package_id = p2.package_id AND d2.type = 0 AND dl2.reqprov = %d)"
"              AND dl.dep_id = dl2.dep_id AND (%s)"
"        GROUP BY p2.name"
"        ORDER BY p2.name DESC", prefix, prefix, prefix, prefix, prefix, prefix, str, !depends, depends, str2);

    g_free(str);
    g_free(str2);

    return result;
}


static gchar* _get_depends_requires_sql(const gchar *name, const gchar *prefix, gboolean depends)
{
    GList *list = NULL;
    gchar *result;

    list = g_list_prepend(list, g_strdup(name));
    result = _get_depends_requires_many_sql(list, prefix, depends);
    g_list_free(list);

    return result;
}

static gchar* _get_depends_many_sql(GList *names, const gchar *prefix)
{
    gchar *str = NULL;
    gchar *result;
    GList *li;
    gchar **str_array = NULL;
    gint n;

    n = g_list_length(names);
    str_array = g_new (gchar*, n + 1);
    str_array[n--] = NULL;

    for (li = names; li != NULL; li = li->next)
        str_array[n--] = g_strdup_printf("r.package = '%s'", (gchar*)li->data);

    str = g_strjoinv(" OR ", str_array);
    g_strfreev(str_array);


    result = g_strdup_printf(""
"        SELECT r.package, p.package, r.dep"
"        FROM require_%s AS r"
"        LEFT JOIN provide_%s AS p ON r.dep = p.dep"
"        WHERE %s", prefix, prefix, str);

    g_free(str);

    return result;
}

#if GENERATE_GRAPH
static gchar* _get_depends_sql(const gchar *name, const gchar *prefix)
{
    GList *list = NULL;
    gchar *result;

    list = g_list_prepend(list, g_strdup(name));
    result = _get_depends_many_sql(list, prefix);
    g_list_free(list);

    return result;
}
#endif

static GList* _box_db_repos_get_depends_requires(sqlite3 *db, const gchar *name, gboolean depends)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *files = NULL;
    gchar *sql = NULL;
    PackageSearch *search = NULL;
    GList *li = NULL;
    gchar *prefix = NULL;
    gint count = 0;

    for (li = box_db_get_repos(); li != NULL; li = li->next)
    {
        prefix = (gchar *)li->data;
        sql = g_strdup_printf("SELECT COUNT(*) FROM %s.package AS p WHERE p.name = '%s'", prefix, name);
        count = box_db_execute_simple_query_get_int(db, sql);
        g_free(sql);
        if (count == 1)
            break;
    }

    if (count != 1)
        return NULL;

    sql = _get_depends_requires_sql(name, prefix, depends);
    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            search = box_db_repos_package_search_new(
                        g_strdup((gchar *)sqlite3_column_text(stmt, 0)), 
                        g_strdup((gchar *)sqlite3_column_text(stmt, 1)),
                        g_strdup((gchar *)sqlite3_column_text(stmt, 2)),
                        g_strdup(DEFAULT_ARCH),
                        TRUE, /* FIXME: installed? */
                        g_strdup(""),
                        g_strdup(prefix),
                        TRUE /* FIXME: gui? */
                   );
            files = g_list_prepend(files, search);
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);
    g_free(sql);

    return files;
}

GList* box_db_repos_get_depends(sqlite3 *db, const gchar *name)
{
    return _box_db_repos_get_depends_requires(db, name, TRUE);
}

GList* box_db_repos_get_requires(sqlite3 *db, const gchar *name)
{
    return _box_db_repos_get_depends_requires(db, name, FALSE);
}

/* TODO: create and use cache - DROP TABLE IF EXISTS cache_deps; CREATE TEMPORARY TABLE cache_deps (package TEXT, dep TEXT); INSERT INTO cache_deps SELECT DISTINCT p.name AS x1, p2.name AS x2 FROM dep AS d JOIN dep_list AS dl ON d.dep_id = dl.dep_id LEFT OUTER JOIN package AS p ON dl.package_id = p.package_id, dep AS d2 JOIN dep_list AS dl2 ON d2.dep_id = dl2.dep_id JOIN package AS p2 ON dl2.package_id = p2.package_id WHERE (d.type = 0 AND dl.reqprov = 0) AND (d2.type = 0 AND dl2.reqprov = 1) AND dl.dep_id = dl2.dep_id AND p2.name != p.name; */

DepsResolve* box_db_repos_resolve_deps(sqlite3 *db, GList *names)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *deps = NULL;
    GList *missing = NULL;
    gchar *sql = NULL;
    GList *li = NULL;
    gchar *prefix = NULL;
    gint count = 0;
    DepsResolve *result;
#if GENERATE_GRAPH
    gchar *name;
#endif

    /* FIXME!!! */
    for (li = box_db_get_repos(); li != NULL; li = li->next)
    {
        prefix = (gchar *)li->data;
        sql = g_strdup_printf("SELECT COUNT(*) FROM %s.package AS p WHERE p.name = '%s'", prefix, (gchar*)g_list_first(names)->data);
        count = box_db_execute_simple_query_get_int(db, sql);
        g_free(sql);
        if (count == 1)
            break;
    }

    if (count != 1)
    {
        result = g_new0(DepsResolve, 1);
        result->deps = NULL;
        result->missing = NULL;

        return result;
    }


#if GENERATE_GRAPH
    for (li = names; li != NULL; li=li->next)
    {
    name = (gchar *)li->data;
    sql = _get_depends_sql(name, prefix);
#else
    sql = _get_depends_many_sql(names, prefix);
#endif
    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            if (sqlite3_column_text(stmt, 1) != NULL)
            {
                if (!g_list_find_custom(deps, sqlite3_column_text(stmt, 1), compare_string_func))
                {
                    deps = g_list_prepend(deps, g_strdup((gchar *)sqlite3_column_text(stmt, 1)) );
#if GENERATE_GRAPH
                    fprintf(stderr, "%s\t%s\n", name, (gchar *)sqlite3_column_text(stmt, 1));
#endif
                }
            } else
            {
                if (!g_list_find_custom(missing, sqlite3_column_text(stmt, 2), compare_string_func))
                {
                    missing = g_list_prepend(missing, g_strdup((gchar *)sqlite3_column_text(stmt, 2)) );
                }
            }
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);
    g_free(sql);

#if GENERATE_GRAPH
    }
#endif

    result = g_new0(DepsResolve, 1);

    result->deps = deps;
    result->missing = missing;

    return result;
}

DepsResolve* box_db_repos_resolve_deps_one(sqlite3 *db, gchar *name)
{
    GList *list = NULL;
    DepsResolve *result;

    list = g_list_prepend(list, g_strdup(name));
    result = box_db_repos_resolve_deps(db, list);
    g_list_free(list);

    return result;

}

GList* box_db_repos_get_packages(sqlite3 *db)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *result = NULL;
    gchar *sql = NULL;

    sql = "SELECT name FROM u_package";
    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            result = g_list_prepend(result, g_strdup((gchar *)sqlite3_column_text(stmt, 0)));
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }
    sqlite3_finalize(stmt);

    return result;
}

void box_db_repos_deps_free(DepsResolve *deps)
{
    GList *li;

    for (li = deps->deps; li != NULL; li = li->next)
        g_free(li->data);
    g_list_free(deps->deps);

    for (li = deps->missing; li != NULL; li = li->next)
        g_free(li->data);
    g_list_free(deps->missing);

    g_free(deps);
}

#if USE_TSORT
static GList* _box_db_repos_parse_tsort_output(gchar *str)
{
    GList *result = NULL;
    gchar **lines = NULL;
    gint i = 0;

    lines = g_strsplit(str, "\n", 0);

    while (lines[i])
    {
        if (strlen(lines[i]) > 0)
        {
            result = g_list_prepend(result, g_strdup(lines[i]));
        }
        i++;
    }

    g_strfreev(lines);

    return result;
}

static GList* _box_db_repos_resolve_deps_recurse_tsort(sqlite3 *db, DepsResolve* deps, gboolean install_mode)
{
    DepsResolve *tmp = NULL;
    GList *result = NULL;
    GList *li, *li2;
    gint file;
    gchar *filename;
    gchar *line;
    gint size;
    GList *list = NULL;
    gchar *cmd_stdout = NULL;
    gchar *cmd_stderr = NULL;
    gchar *cmdline = NULL;

    file = g_file_open_tmp("box-deps-XXXXXX", &filename, NULL);

    /* result contains unique dependencies in unsorted order, we need to sort it */
    for (li = deps->deps; li != NULL; li = li->next)
    {
        gint subdeps = 0;

        if (install_mode)
        {
            /* skip installed packages in install mode */
            if (box_db_get_package_id(db, li->data) >= 0)
                continue;
        }

        list = g_list_prepend(list, g_strdup(li->data));
        tmp = box_db_repos_resolve_deps(db, list);
        g_list_free(list);
        list = NULL;

        for (li2 = tmp->deps; li2 != NULL; li2 = li2->next)
        {
            if (strcmp((char *)li->data, (char *)li2->data) == 0)
                continue;

            if (install_mode)
            {
                /* skip installed packages in install mode */
                if (box_db_get_package_id(db, li2->data) >= 0)
                    continue;
            }

            line = g_strdup_printf("%s %s\n", (char *)li->data, (char *)li2->data);
            size = write(file, line, strlen(line));
            g_free(line);
            
            subdeps++;
        }
        if (subdeps == 0)
        {
            line = g_strdup_printf("%s %s\n", (char *)li->data, (char *)li->data);
            size = write(file, line, strlen(line));
            g_free(line);
        }
        box_db_repos_deps_free(tmp);
    }

    close(file);

    /* use external tsort command */
    cmdline = g_strdup_printf("tsort %s", filename);
    g_spawn_command_line_sync(cmdline, &cmd_stdout, &cmd_stderr, NULL, NULL);
    g_free(cmdline);

    result = _box_db_repos_parse_tsort_output(cmd_stdout);

    g_unlink(filename);

    return result;
}
#endif


DepsResolve* box_db_repos_resolve_deps_recurse(sqlite3 *db, const gchar *name, gboolean install_mode)
{
    DepsResolve *result = NULL;
    DepsResolve *deps = NULL;
    GList *li;
    GList *current = NULL;
    gint added = 0;
#if USE_TSORT
    GList *sorted_deps = NULL;
#endif

    current = g_list_prepend(current, (gchar *)name);

    result = g_new0(DepsResolve, 1);

    while (1)
    {
        deps = box_db_repos_resolve_deps(db, current);
        g_list_free(current);
        current = NULL;

        added = 0;
        for (li = deps->deps; li != NULL; li = li->next)
        {
            /* append new results only */
            if (!g_list_find_custom(result->deps, li->data, compare_string_func) &&
                strcmp(name, li->data) != 0)
            {
                if (install_mode)
                {
                    /* if package not installed */
                    if (box_db_get_package_id(db, li->data) < 0) {
                        current = g_list_append(current, g_strdup(li->data));
                        result->deps = g_list_prepend(result->deps, g_strdup(li->data));
                        added++;
                    }
                } else {
                    current = g_list_append(current, g_strdup(li->data));
                    result->deps = g_list_prepend(result->deps, g_strdup(li->data));
                    added++;
                }
            }
        }

        /* collect broken dependencies */
        for (li = deps->missing; li != NULL; li = li->next)
        {
            /* append new results only */
            if (!g_list_find_custom(result->missing, li->data, compare_string_func))
            {
                result->missing = g_list_prepend(result->missing, g_strdup(li->data));
            }
        }
        box_db_repos_deps_free(deps);

        if (added == 0) {
            g_list_free(current);
            break;
        }
    }

#if USE_TSORT
    sorted_deps = _box_db_repos_resolve_deps_recurse_tsort(db, result, install_mode);

    /* free unsorted deps */
    for (li = result->deps; li != NULL; li = li->next)
        g_free(li->data);
    g_list_free(result->deps);

    /* assign sorted deps */
    result->deps = sorted_deps;
#endif

    return result;
}
