#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sqlite3.h>

#include "libbox-db.h"
#include "libbox.h"

#define TABLE_PACKAGE "\
CREATE TABLE package(\n\
    'package_id' INTEGER PRIMARY KEY,\n\
    'name' TEXT,\n\
    'version' TEXT,\n\
    'release' INTEGER,\n\
    'inst_date' TEXT,\n\
    'summary' TEXT)"

#define TABLE_DEP "\
CREATE TABLE dep(\n\
    'dep_id' INTEGER PRIMARY KEY,\n\
    'name' TEXT,\n\
    'type' INTEGER)"

#define TABLE_DEP_LIST "\
CREATE TABLE dep_list(\n\
    'dep_id' INTEGER,\n\
    'package_id' INTEGER,\n\
    'reqprov' INTEGER,\n\
    'state' TEXT)"
    
#define TABLE_FILE_LIST "\
CREATE TABLE file_list(\n\
    'package_id' INTEGER,\n\
    'file_id' INTEGER)"

#define TABLE_FILE "\
CREATE TABLE file(\n\
    'file_id' INTEGER PRIMARY KEY,\n\
    'parent_id' INTEGER,\n\
    'name' TEXT,\n\
    'symlink' TEXT,\n\
    'type' TEXT,\n\
    'perms' TEXT,\n\
    'hash' TEXT)"

#define TABLE_META "\
CREATE TABLE meta(\n\
    'meta_id' INTEGER PRIMARY KEY,\n\
    'name' TEXT,\n\
    'value' TEXT)"

#define TABLE_SCRIPT "\
CREATE TABLE script(\n\
    'package_id' INTEGER,\n\
    'type' INTEGER,\n\
    'content' TEXT,\n\
    'status' INTEGER)"

#define INDEXES "\
    CREATE INDEX idx1 ON file(name); \
    CREATE INDEX idx2 on file(parent_id); \
    CREATE INDEX idx3 ON file_list(file_id);"


sqlite3* box_db_open(const gchar *path)
{
    sqlite3 *db = NULL;
    gchar *dbfile;
    gchar *file;
    gchar *dir;

    dbfile = box_get_db_file();
    file = g_build_filename(path, dbfile, NULL);
    dir = g_path_get_dirname(file);

    g_free(dbfile);
    /* Ensure directory is created for DB file */
    g_mkdir_with_parents(dir, 0755);

    g_free(dir);

    /* Create DB here */
    sqlite3_open(file, &db);
    g_free(file);
    
    return db;
}

void box_db_close(sqlite3 *db)
{
    if (db != NULL)
        sqlite3_close(db);
}

int box_db_execute_simple_query_get_int(sqlite3 *db, gchar *sql )
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    int result = -1;
    int counter = 0;

    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            result = sqlite3_column_int(stmt, 0);
            counter++;
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }

    sqlite3_finalize(stmt);

    g_assert(counter == 0 || counter == 1);

    return result;
}

gchar* box_db_execute_simple_query_get_string(sqlite3 *db, gchar *sql )
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    gchar* result = NULL;
    int counter = 0;

    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            result = g_strdup((gchar *)sqlite3_column_text(stmt, 0));
            counter++;
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }

    sqlite3_finalize(stmt);

    g_assert(counter == 0 || counter == 1);

    return result;
}

GList*
box_db_execute_simple_query_get_string_list(sqlite3 *db, gchar *sql)
{
    int rc;
    sqlite3_stmt *stmt = NULL;
    GList *result = NULL;

    rc = sqlite3_prepare(db, sql, strlen(sql), &stmt, NULL);

    while ( (rc = sqlite3_step(stmt)) == SQLITE_ROW )
    {
        if (rc == SQLITE_ROW)
        {
            result = g_list_prepend(result, g_strdup((gchar *)sqlite3_column_text(stmt, 0)));
        } else if (rc != SQLITE_OK)
        {
            printf("ERROR: %s\n", sqlite3_errmsg(db));
        }
    }

    sqlite3_finalize(stmt);

    return result;
}


void box_db_transaction_rollback(sqlite3 *db)
{
    sqlite3_exec(db, "ROLLBACK", NULL, NULL, NULL);
}

void box_db_show_exec_error(sqlite3 *db, char **errmsg, gboolean fatal)
{
    if (*errmsg)
    {
        printf("Error: %s\n", *errmsg);
        sqlite3_free(*errmsg);
        if (fatal)
        {
            box_db_transaction_rollback(db);
            exit(1);
        }
    }
}

static void box_db_execute_simple_light(sqlite3 *db, gchar *sql)
{
    char *errmsg = NULL;
    sqlite3_exec(db, sql, NULL, NULL, &errmsg);
/*    box_db_show_exec_error(db, &errmsg, FALSE); */
    sqlite3_free(errmsg);
}

void box_db_execute_simple(sqlite3 *db, gchar *sql)
{
    char *errmsg = NULL;
    sqlite3_exec(db, sql, NULL, NULL, &errmsg);
    box_db_show_exec_error(db, &errmsg, TRUE);
}

void transaction_rollback(void* connection)
{
    printf("ERROR: transaction has been rolled back!\n");
}

void box_db_transaction_begin(sqlite3 *db)
{
    sqlite3_rollback_hook(db, transaction_rollback, NULL);
    box_db_execute_simple(db, "BEGIN");
}

void box_db_transaction_end(sqlite3 *db)
{
    box_db_execute_simple(db, "COMMIT");
}


// TODO:
static int box_db_migration_version_get(sqlite3 *db)
{
    return 0;
}

// TODO:
static void box_db_migration_version_set(sqlite3 *db, int version)
{
}

static void box_db_migration_to(sqlite3 *db, int version)
{
    switch (version)
    {
        case 1:
            
        break;

        default:
        break;
    }
}

static void box_db_migration_run(sqlite3 *db)
{
    int current = 0;
    int version;
    int i;

    // 1. read current version
    version = box_db_migration_version_get(db);

    // 2. migrate
    for (i = version; i < current; i++)
    {
        box_db_migration_to(db, i + 1);
    }

    // 3. update version
    box_db_migration_version_set(db, current);
}

void box_db_create_empty(sqlite3 *db)
{
    box_db_execute_simple_light(db, TABLE_PACKAGE);
    box_db_execute_simple_light(db, TABLE_DEP);
    box_db_execute_simple_light(db, TABLE_DEP_LIST);
    box_db_execute_simple_light(db, TABLE_FILE_LIST);
    box_db_execute_simple_light(db, TABLE_FILE);
    box_db_execute_simple_light(db, TABLE_META);
    box_db_execute_simple_light(db, TABLE_SCRIPT);
    box_db_execute_simple_light(db, INDEXES);

    box_db_migration_run(db);
}

