#include <libxml/xmlmemory.h>
#include <libxml/xmlwriter.h>
#include <libxml/parser.h>

#include "libbox-parser.h"


static void _box_header_parse_name(xmlDocPtr doc, xmlNodePtr cur, BoxHeader *header)
{
    xmlChar *version = xmlGetProp(cur, (const xmlChar *) "version");
    xmlChar *release = xmlGetProp(cur, (const xmlChar *) "release");
 
    if (version != NULL)
    {
        header->version = g_strdup((gchar *)version);
        xmlFree(version);
    }

    if (release != NULL)
    {
        header->release = g_strdup((gchar *)release);
        xmlFree(release);
    }


    cur = cur->xmlChildrenNode;
    header->name = g_strdup((gchar *)cur->content);
}

static void _box_header_parse_source(xmlDocPtr doc, xmlNodePtr cur, BoxHeader *header)
{
    xmlChar *group = xmlGetProp(cur, (const xmlChar *) "id");

    if (xmlStrcmp(group, (const xmlChar *) "main") == 0)
    {

        cur = cur->xmlChildrenNode;
        while (cur != NULL)
        {
            if ((xmlStrcmp(cur->name, (const xmlChar *) "text")))
            {

                if (xmlStrcmp(cur->name, (const xmlChar *) "name") == 0)
                {
                    _box_header_parse_name(doc, cur, header);
                }
                /*
                 * TODO: read url, homepage
                 */
            }
            cur = cur->next;
        }
    }
    xmlFree(group);
}

static void _box_header_parse_sources(xmlDocPtr doc, xmlNodePtr cur, BoxHeader *header)
{
    cur = cur->xmlChildrenNode;
    while (cur != NULL)
    {
        if ((xmlStrcmp(cur->name, (const xmlChar *) "text")))
        {
            if (xmlStrcmp(cur->name, (const xmlChar *) "source") == 0)
            {
                _box_header_parse_source(doc, cur, header);
            }
        }
        cur = cur->next;
    }
}

static BoxHeader* _box_header(xmlDocPtr doc, xmlNodePtr cur)
{
    BoxHeader *header = NULL;
    xmlChar *group;

    header = g_new0(BoxHeader, 1);

    group = xmlGetProp(cur, (const xmlChar *) "group");
    
    if (group != NULL)
    {
        header->group = g_strdup((gchar *)group);
        xmlFree(group);
    }

    _box_header_parse_sources(doc, cur, header);


    if (header->name == NULL)
        header->name = g_strdup("");

    if (header->version == NULL)
        header->version = g_strdup("");

    if (header->release == NULL)
        header->release = g_strdup("");

    if (header->group == NULL)
        header->group = g_strdup("");

    return header;
}

static gchar* _box_packages_parse_package_script(xmlDocPtr doc, xmlNodePtr cur, const gchar* tag_name)
{
    gchar *script = NULL;

    cur = cur->xmlChildrenNode;
    while (cur != NULL)
    {
        if ((xmlStrcmp(cur->name, (const xmlChar *) "text")))
        {

            if (xmlStrcmp(cur->name, (const xmlChar *) tag_name) == 0)
            {
                script = g_strdup((gchar *)cur->xmlChildrenNode->content);
            }
        }
        cur = cur->next;
    }

    return script;
}

static BoxPackage* _box_packages_parse_package(xmlDocPtr doc, xmlNodePtr cur, const gchar* pkg_name)
{
    BoxPackage *package = NULL;
    xmlChar *name;
    gchar *gname;

    name = xmlGetProp(cur, (const xmlChar *) "name");

    if (name == NULL)
        gname = g_strdup(pkg_name);
    else
        gname = g_strdup((gchar *)name);

    xmlFree(name);

    package = g_new0(BoxPackage, 1);

    package->name = gname;
    package->pre_install = _box_packages_parse_package_script(doc, cur, "pre-install");
    package->post_install = _box_packages_parse_package_script(doc, cur, "post-install");

    return package;
}

static GList* _box_packages(xmlDocPtr doc, xmlNodePtr cur, const gchar *name)
{
    GList *packages = NULL;
    BoxPackage *package = NULL;

    cur = cur->xmlChildrenNode;
    while (cur != NULL)
    {
        if ((xmlStrcmp(cur->name, (const xmlChar *) "text")))
        {
            if (xmlStrcmp(cur->name, (const xmlChar *) "package") == 0)
            {
                package = _box_packages_parse_package(doc, cur, name);
                if (package)
                    packages = g_list_append(packages, package);
            }
        }
        cur = cur->next;
    }


    return packages;
}

/*/////////////////////////////////////////// */

BoxDocument* box_document_new(const gchar *filename)
{
    xmlDocPtr doc = NULL;
    xmlNodePtr cur;
    BoxDocument *document = NULL;

    doc = xmlParseFile(filename);

    if (doc == NULL)
    {
        fprintf(stderr, "Document not parsed successfully.\n");
        return NULL;
    }

    cur = xmlDocGetRootElement(doc);

    if (cur == NULL)
    {
        fprintf(stderr, "empty document\n");
        xmlFreeDoc(doc);

        return NULL;
    }

    if (xmlStrcmp(cur->name, (const xmlChar *) "box"))
    {
        fprintf(stderr, "document of the wrong type, root node != box\n");
        xmlFreeDoc(doc);

        return NULL;
    }

    document = g_new0(BoxDocument, 1);
    document->header = _box_header(doc, cur);
    document->packages = _box_packages(doc, cur, document->header->name);

    xmlFreeDoc(doc);

    return document;
}

void box_document_free(BoxDocument *document)
{
    g_free(document->header->name);
    g_free(document->header->version);
    g_free(document->header->release);
    g_free(document->header->group);

    g_free(document->header);

    g_free(document);
}
