#include <glib.h>
#include <stdio.h>
#include <string.h>

#include "config.h"
#include "libbox-repos.h"
#include "libbox.h"
#include "libbox-db.h"
#include "libbox-download.h"
#include "libbox-db-utils.h"

RepoInfo* box_repos_repo_new(const gchar *name, const gchar *description, const gchar *url, gboolean enabled)
{
    RepoInfo *repo = g_new0(RepoInfo, 1);

    repo->name = g_strdup(name);
    repo->description = g_strdup(description);
    repo->url = g_strdup(url);
    repo->enabled = enabled;

    return repo;
}

void box_repos_repo_free(RepoInfo *repo)
{
    g_free(repo->name);
    g_free(repo->description);
    g_free(repo->url);
    g_free(repo);
}


GList* box_repos_list_get()
{
    GKeyFile *key_file;
    gchar **repos = NULL;
    gsize count;
    gint i;
    GList *result = NULL;
    RepoInfo *repo = NULL;
    gchar *desc, *url, *repos_file;
    gboolean enabled;

    repos_file = box_get_repos_conf();
    key_file =  g_key_file_new();
    if (g_key_file_load_from_file(key_file, repos_file, G_KEY_FILE_NONE|G_KEY_FILE_KEEP_COMMENTS, NULL))
    {
        repos = g_key_file_get_groups(key_file, &count);
        for (i=0; i<count; i++)
        {
            desc = g_key_file_get_string(key_file, repos[i], "description", NULL);
            url = g_key_file_get_string(key_file, repos[i], "url", NULL);
            enabled = g_key_file_get_boolean(key_file, repos[i], "enabled", NULL);

            repo = box_repos_repo_new(repos[i], desc, url, enabled);

            g_free(desc);
            g_free(url);

            result = g_list_append(result, repo);
        }
        g_strfreev(repos);
    }
    g_free(repos_file);

    g_key_file_free(key_file);

    return result;
}

void box_repos_list_free(GList *list)
{
    GList *li;

    for (li = list; li != NULL; li = li->next)
    {
        box_repos_repo_free((RepoInfo*)li->data);
    }
    g_list_free(list);
}

gboolean box_repos_enable_repo(const gchar *name, gboolean enable)
{
    GKeyFile *key_file;
    gboolean result = FALSE;
    gchar *filestr, *repos_file;

    repos_file = box_get_repos_conf();
    key_file = g_key_file_new();
    if (g_key_file_load_from_file(key_file, repos_file, G_KEY_FILE_NONE|G_KEY_FILE_KEEP_COMMENTS, NULL))
    {
        if (g_key_file_has_group(key_file, name))
        {
            if (g_key_file_get_boolean(key_file, name, "enabled", NULL) != enable)
            {
                g_key_file_set_boolean(key_file, name, "enabled", enable);
                result = TRUE;
            }

        }
    }

    filestr = g_key_file_to_data(key_file, NULL, NULL);
    if (!g_file_set_contents(repos_file, filestr, -1, NULL))
        result = FALSE;

    g_free(repos_file);
    g_free(filestr);
    g_key_file_free(key_file);

    return result;
}

gboolean box_repos_set_param(const gchar *group, const gchar *key, const gchar *value)
{
    GKeyFile *key_file;
    gboolean result = FALSE;
    gchar *filestr, *repos_file;
    gboolean valid_key = FALSE;
    gint i;
    const gchar *valid_keys[] = { "description", "enabled", "url" };

    for (i=0; i<3; i++)
    {
	    if (strcmp(key, valid_keys[i]) == 0)
	    {
		    valid_key = TRUE;
		    break;
	    }
    }

    if (!valid_key)
	    return FALSE;

    repos_file = box_get_repos_conf();
    key_file = g_key_file_new();
    if (g_key_file_load_from_file(key_file, repos_file, G_KEY_FILE_NONE|G_KEY_FILE_KEEP_COMMENTS, NULL))
    {
        if (g_key_file_has_group(key_file, group))
        {
	    g_key_file_set_value(key_file, group, key, value);
	    result = TRUE;
        }
    }

    filestr = g_key_file_to_data(key_file, NULL, NULL);
    if (!g_file_set_contents(repos_file, filestr, -1, NULL))
        result = FALSE;

    g_free(repos_file);
    g_free(filestr);
    g_key_file_free(key_file);

    return result;
}

void _box_repos_set_location(gchar *dbfile, gchar *location)
{
    sqlite3 *db = NULL;
    gchar *sql = NULL;

    sqlite3_open(dbfile, &db);

    sql = g_strdup_printf("UPDATE meta SET value = '%s' WHERE name = 'location'", location);

    box_db_execute_simple(db, sql);

    g_free(sql);

    if (db != NULL)
        sqlite3_close(db);
}

void box_repos_sync(const gchar *dest_dir, progressFunc progress, gpointer user_data)
{
    GList *repos;
    GList *li;
    RepoInfo *repo;
    gchar *file, *dbdir;
    gchar *out, *out2;

#define FILE_PROTO "file://"
#define HTTP_PROTO "http://"

    repos = box_repos_list_get();

    if (progress)
        progress(0, user_data);

    dbdir = box_get_db_dir();
    for (li = repos; li != NULL; li=li->next)
    {
        repo = (RepoInfo*) li->data;
        if (repo->enabled)
        {
            if (g_str_has_prefix(repo->url, FILE_PROTO) || g_str_has_prefix(repo->url, HTTP_PROTO))
            {
                file = g_build_filename(repo->url, "index", NULL);
                out2 = g_strdup_printf("index-%s", repo->name);
                out = g_build_filename(dest_dir, dbdir, out2, NULL);
                g_free(out2);

                printf("Downloading repository index %s.\nPlease wait...\n", repo->url);
                if (!box_download_file(file, out)) {
                    printf("Error: cannot refresh repository '%s'.\n", repo->name);
                } else {
                    /* set location in the index */
                    _box_repos_set_location(out, repo->url);
                }

                g_free(file);
                g_free(out);
            } else
            {
                printf("URL: '%s' is not supported\n", repo->url);
            }
        }
    }
    g_free(dbdir);

    box_repos_list_free(repos);

    if (progress)
        progress(100, user_data);
}

